#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "mdbtool.h"

int main (int argc, char *argv[]) {
	const char *command;
	STR *object;
	
	command = argv[1];
	object = (STR*)argv[2];
	
	if ((command == NULL)||(object == NULL)) {
		printf("Usage: mdbtool <command> <object> ...\n");
		return 0;
	}
	
	if (!MemoryManagerOpen("mdbtool")) {
		printf("ERROR: Could not start memory manager\n");
		return 1;
	}
	
	if (!MDBInit()) {
		printf("ERROR: Could not initialise MDB\n");
		return 1;
	}
	
	MDBHandle *h = MDBAuthenticate((STR*)"Hula", NULL, NULL);
	if (h == NULL) {
		printf("ERROR: Could not authenticate to MDB\n");
		return 1;
	}
	
	if (strncmp(command, "getattr", 7) == 0) {
		STR *attrname = (STR*)argv[3];
		if (attrname == NULL) {
			printf("Usage: mdbtool getattr <object> <attrname>\n");
			return 0;
		}
		GetAttr(h, object, attrname);
	} else if (strncmp(command, "setattr", 7) == 0) {
		STR *attrname = (STR*)argv[3];
		STR *value = (STR*)argv[4];
		if ((attrname == NULL) || (value == NULL)) {
			printf("Usage: mdbtool setattr <object> <attrname> <value>\n");
			return 0;
		}
		if (SetAttr(h, object, attrname, value)) {
			printf("ERROR: Could not write attribute\n");
			return 1;
		}
	} else if (strncmp(command, "listattr", 8) == 0) {
		if (object == NULL) {
			printf("Usage: mdbtool listattr <object>\n");
			return 0;
		}
		ListAttr(h, object);
	} else if (strncmp(command, "addobject", 9) == 0) {
		STR *type = (STR*)argv[3];
		if (type == NULL) {
			printf("Usage: mdbtool addobject <object> <class> [attr1=value1 attr2=...]\n");
			return 0;
		}
		struct dict *attrs = GetDict(&argv[4], argc-4);
		if (AddObject(h, object, type, attrs)) {
			printf("ERROR: Could not add object\n");
			return 1;
		}
	} else if (strncmp(command, "delobject", 9) == 0) {
		BOOL recurse = 0;
		if (argv[3] && (strncmp(argv[3], "recurse", 7) == 0)) {
			recurse = 1;
		}
		if (DelObject(h, object, recurse) == 0) {
			/* FIXME: seems this happens when we recurse; object 
			removed fine when command re-run. Maybe DelObject 
			returns false even when it worked, when recursing? */
			printf("ERROR: Could not remove object\n");
			return 1;
		}
	} else if (strncmp(command, "setpassword", 11) == 0) {
		STR *password = (STR*)argv[3];
		if (SetPassword(h, object, password)) {
			printf("ERROR: Could not change the password.\n");
			return 1;
		}
	} else if (strncmp(command, "gettype", 13) == 0) {
		if (object == NULL) {
			printf("Usage: mdbtool gettype <object>\n");
			return 0;
		} 
		GetType(h, object);
	} else if (strncmp(command, "getchildren", 11) == 0) {
		const STR *type = (STR*)argv[3];
		const STR *filter = (STR*)argv[4];
		if (object == NULL) {
			printf("Usage: mdbtool getchildren <object> <type> <filter>\n");
			return 0;
		}
		GetChildren(h, object, type, filter);
	} else {
		printf("Unsupported command %s.\n", command);
	}	
	
	return 0;
}

void GetAttr(MDBHandle *h, const STR *object, const STR *attr) {
	MDBValueStruct *vs;
	MDBEnumStruct *es;
	const unsigned char *attrval;
	vs = MDBCreateValueStruct(h, NULL);
	es = MDBCreateEnumStruct(vs);
	
	while ((attrval = MDBReadEx(object, attr, es, vs)) != NULL) {
		printf ("%s::%s\n", attr, attrval);
	}
	
	MDBDestroyEnumStruct(es, vs);
	MDBDestroyValueStruct(vs);
}

BOOL SetAttr(MDBHandle *h, const STR *object, const STR *attr, const STR *value) {
	MDBValueStruct *vs;
	BOOL result;
	
	vs = MDBCreateValueStruct(h, object);
	MDBAddValue(value, vs);
	result = MDBWrite(object, attr, vs);
	MDBDestroyValueStruct(vs);
	
	return result;
}

BOOL AddObject(MDBHandle *h, const STR *object, const STR *class, struct dict *attrs) {
	MDBValueStruct *attr, *value, *vs;
	BOOL result;
	
	attr = MDBCreateValueStruct(h, object);
	value = MDBCreateValueStruct(h, object);
	vs = MDBCreateValueStruct(h, object);
	
	while (attrs != NULL) {
		if ((attrs->key != NULL) && (attrs->value != NULL)) {
			MDBAddValue(attrs->key, attr);
			MDBAddValue(attrs->value, value);
		}
		attrs = attrs->next;
	}
	
	result = MDBCreateObject(object, class, attr, value, vs);
	
	MDBDestroyValueStruct(vs);
	MDBDestroyValueStruct(value);
	MDBDestroyValueStruct(attr);
	return (!result);
}

BOOL DelObject(MDBHandle *h, const STR *object, BOOL recurse) {
	BOOL result;
	MDBValueStruct *vs = MDBCreateValueStruct(h, object);
	
	result = MDBDeleteObject(object, recurse, vs);
	
	MDBDestroyValueStruct(vs);
	return result;
}

BOOL GetType(MDBHandle *h, const STR *object) {
	BOOL result;
	MDBValueStruct *vs = MDBCreateValueStruct(h, object);
	STR type[1024];
	
	result = MDBGetObjectDetails(object, (STR *)&type, NULL, NULL, vs);
	printf("type:%s\n", type);
	
	MDBDestroyValueStruct(vs);
	return result;
}

BOOL GetChildren(MDBHandle *h, const STR *object, const STR *type, const STR *filter) {
	BOOL result;
	MDBValueStruct *vs = MDBCreateValueStruct(h, object);
	MDBEnumStruct *es = MDBCreateEnumStruct(vs);
	const STR *child;
	
	/* FIXME: filter does not seem to work. Hmmmm :( */
	while((child = MDBEnumerateObjectsEx(object, type, filter, 0, es, vs)) != NULL) {
		printf("child:%s\n", child);
	}
	
	MDBDestroyEnumStruct(es, vs);
	MDBDestroyValueStruct(vs);
	return result;
}

BOOL SetPassword(MDBHandle *h, const STR *object, const STR *password) {
	BOOL result;
	MDBValueStruct *vs = MDBCreateValueStruct(h, object);
	
	result = MDBChangePasswordEx(object, NULL, password, vs);
	
	MDBDestroyValueStruct(vs);
	return (!result);
}

struct dict *GetDict(char *list[], int items) {
	struct dict *d;
	char *item;
	
	if (items == 0) return NULL;
	item = list[0];
	
	d = malloc(sizeof(struct dict));
	d->key = (STR *)item;
	d->value = (STR *)strchr(item, '=');
	*d->value = 0;
	d->value++;
	d->next = GetDict(&list[1], items - 1);
	
	return d;
}

/* broken vv */

void ListAttr(MDBHandle *h, const STR *object) {
	printf("ERROR: This function is currently broken, sorry :(\n");
	return;
/*
	MDBValueStruct *vs = MDBCreateValueStruct(handle, object);
	MDBEnumStruct *es = MDBCreateEnumStruct(vs);
	
	const STR *attr;
	
	while ((attr = MDBEnumerateAttributesEx(object, es, vs)) != NULL) {
		long l = MDBRead(object, attr, vs);
		if (l != 0) {
			long i;
			for (i=0; i<l; i++) {
				printf("%s: %s\n", attr, vs->Value[i]);
			}
		}
	}
	
	MDBDestroyEnumStruct(es, vs);
	MDBDestroyValueStruct(vs);
*/
}
